import re

from django.utils.formats import localize_input
from django.template.defaultfilters import register


def do_linebreaks(value: str) -> str:
    return value.replace("\n", "\\\\\n")


REPLACEMENTS = dict(
    [
        ("&", "\\&"),
        ("%", "\\%"),
        ("$", "\\$"),
        ("#", "\\#"),
        ("_", "\\_"),
        ("{", "\\{"),
        ("}", "\\}"),
        ("\\", "\\textbackslash{}"),
        ("~", "\\textasciitilde{}"),
        ("^", "\\textasciicircum{}"),
    ]
)

ESCAPE_PATTERN = re.compile("[{}]".format("".join(map(re.escape, REPLACEMENTS.keys()))))


def do_latex_escape(value: str) -> str:
    """
    Replace all LaTeX characters that could cause the latex compiler to fail
    and at the same time try to display the character as intended from the user.

    see also https://tex.stackexchange.com/questions/34580/escape-character-in-latex
    """
    return ESCAPE_PATTERN.sub(lambda mo: REPLACEMENTS.get(mo.group()), value)


tex_specific_filters = {
    "localize": localize_input,
    "linebreaks": do_linebreaks,
    "latex_escape": do_latex_escape,
}

FILTERS = register.filters.copy()
FILTERS.update(tex_specific_filters)



from django.http import HttpResponse


class PDFResponse(HttpResponse):
    def __init__(self, content, filename=None):
        super(PDFResponse, self).__init__(content_type="application/pdf")
        self["Content-Disposition"] = 'filename="{}"'.format(filename)
        self.write(content)



import datetime
from decimal import Decimal

from django.test import TestCase
from django.test.utils import override_settings
from django.core.exceptions import ValidationError
from django.http import HttpResponse
from django.template import engines

from django_tex.core import (
    run_tex,
    compile_template_to_pdf,
    render_template_with_context,
)
from django_tex.exceptions import TexError

from django_tex.shortcuts import render_to_pdf

from .models import TemplateFile


class RunningTex(TestCase):
    """
    Tests calling latex compiler
    """

    def test_run_tex(self):
        """Call different LaTex interpreters with very simple template"""
        source = "\
        \\documentclass{article}\n\
        \\begin{document}\n\
        This is a test!\n\
        \\end{document}"
        interpreters = ["pdflatex", "latexmk -pdf", "lualatex"]
        for name in interpreters:
            with self.subTest(name=name):
                with self.settings(LATEX_INTERPRETER=name):
                    pdf = run_tex(source)
                    self.assertIsNotNone(pdf)

    @override_settings(LATEX_INTERPRETER="does_not_exist")
    def test_wrong_latex_interpreter(self):
        """Using an unknown interpreter raises an Exception"""
        source = "\
        \\documentclass{article}\n\
        \\begin{document}\n\
        This is a test!\n\
        \\end{document}"

        with self.assertRaises(Exception):
            run_tex(source)  # should raise


class Exceptions(TestCase):
    """
    Tests custom exceptions
    """

    def test_exception_emergency_stop(self):
        source = "\
        \\documentclass{article}\n\
        \\begin{document}\n\
        This is a test!\n"

        with self.assertRaises(TexError) as cm:
            run_tex(source)

        self.assertRegex(cm.exception.log, r"^This is Lua")
        self.assertRegex(cm.exception.message, r"^! Emergency stop")
        self.assertRegex(
            cm.exception.message,
            r"(End of file on the terminal\!$)|(job aborted, no legal \\end found)",
        )  # First alternative applies
        # if tex source is given to
        # pdflatex via stdin.
        # Second, if tex source is
        # given as filename

    @override_settings(LATEX_INTERPRETER="pdflatex")
    def test_pdflatex_exceptions(self):
        source = "\
        \\documentclass{article}\n\
        \\begin{document}\n\
        This is a test!\n"

        with self.assertRaises(TexError) as cm:
            run_tex(source)

        self.assertRegex(cm.exception.log, r"^This is pdf")
        self.assertRegex(cm.exception.message, r"^! Emergency stop")
        self.assertRegex(
            cm.exception.message,
            r"(End of file on the terminal\!$)|(job aborted, no legal \\end found)",
        )  # First alternative applies
        # if tex source is given to
        # pdflatex via stdin.
        # Second, if tex source is
        # given as filename

    def test_exception_unknown_command(self):
        source = "\
        \\documentclass{article}\n\
        \\begin{document}\n\
        \\unknown{command}\n\
        \\end{document}\n"

        with self.assertRaises(TexError) as cm:
            run_tex(source)

        self.assertRegex(cm.exception.log, r"^This is Lua")
        self.assertRegex(cm.exception.message, r"^! Undefined control sequence")
        self.assertRegex(cm.exception.message, r"l\.3")

    def test_template_debug(self):
        source = (
            "\\documentclass{article}\n"
            "\\begin{document}\n"
            "\\unknown{command}\n"
            "\\end{document}\n"
        )

        with self.assertRaises(TexError) as cm:
            run_tex(source)

        template_debug = cm.exception.template_debug

        self.assertEqual(template_debug["during"], "\\unknown{command}")
        self.assertEqual(template_debug["line"], 3)

    def test_template_error_context(self):
        source = (
            "\\documentclass{article}\n"
            "\n"
            "\n"
            "\n"
            "\n"
            "\\begin{document}\n"
            "\\unknown{command}\n"
            "\n"
            "\n"
            "\\end{document}\n"
        )

        with self.assertRaises(TexError) as cm:
            run_tex(source)

        message = cm.exception.message

        expected_context = (
            " 2 \n"
            " 3 \n"
            " 4 \n"
            " 5 \n"
            " 6 \\begin{document}\n"
            " 7 \\unknown{command}\n"
            " 8 \n"
            " 9 \n"
            "10 \\end{document}"
        )

        self.assertIn(expected_context, message)


class RenderingTemplates(TestCase):
    """
    Tests rendering a template file with context to a string

    TODO: Add a test for custom template file locations.
    """

    def test_render_template(self):
        template_name = "tests/test.tex"
        context = {
            "test": "a simple test",
            "number": Decimal("1000.10"),
            "date": datetime.date(2017, 10, 25),
            "names": ["Arjen", "Robert", "Mats"],
        }
        output = render_template_with_context(template_name, context)
        self.assertIn("\\section{a simple test}", output)
        self.assertIn("This is a number: 1000,10.", output)
        self.assertIn("And this is a date: 25.10.2017.", output)
        self.assertIn("\\item Arjen", output)

    def test_render_template_from_custom_directory(self):
        template_name = "custom_directory_test.tex"
        context = {"foo": "bar"}
        output = render_template_with_context(template_name, context)
        self.assertIn("bar", output)


class CompilingTemplates(TestCase):
    """
    Tests compiling a template file with a context to a pdf file
    """

    def test_compile_template_to_pdf(self):
        """test compile_template_to_pdf

        - accepts template name and context
        - context may contain unicode characters
        - produces pdf file
        """
        template_name = "tests/test.tex"
        context = {
            "test": "a simple test",
            "number": Decimal("1000.10"),
            "date": datetime.date(2017, 10, 25),
            "names": ["Arjen", "Robert", "Mats", "äüößéèô♞Ⅷ"],
        }
        pdf = compile_template_to_pdf(template_name, context)
        self.assertIsNotNone(pdf)

    @override_settings(LATEX_INTERPRETER="pdflatex")
    def test_compile_template_with_graphics_pdflatex(self):
        template_name = "tests/test_graphics.tex"
        context = {}
        pdf = compile_template_to_pdf(template_name, context)
        self.assertIsNotNone(pdf)


class TemplateLanguage(TestCase):
    """
    Tests features such as whitespace control and filters
    """

    def render_template(self, template_string, context={}, using="tex"):
        engine = engines[using]
        template = engine.from_string(template_string)
        return template.render(context)

    def test_whitespace_control(self):
        context = {"foo": "bar"}
        template_string = "\\section{ {{- foo -}} }"
        output = self.render_template(template_string, context)
        self.assertEqual(output, "\\section{bar}")

    def test_localization(self):
        template_string = "{{ foo|localize }}"
        parameters = [
            ("en", Decimal("1000.10"), "1000.10"),
            ("de-de", Decimal("1000.10"), "1000,10"),
            ("de-de", datetime.date(2017, 10, 25), "25.10.2017"),
        ]
        for lang, value, expected in parameters:
            with self.subTest(lang=lang, value=value):
                with self.settings(LANGUAGE_CODE=lang):
                    output = self.render_template(template_string, {"foo": value})
                    self.assertEqual(output, expected)

    @override_settings(LANGUAGE_CODE="de-de")
    def test_format_long_date(self):
        context = {"foo": datetime.date(2017, 10, 25)}
        template_string = "{{ foo | date('d. F Y') }}"
        # template_string="{{ '{:%d. %B %Y}'.format(foo) }}"
        output = self.render_template(template_string, context)
        self.assertEqual(output, "25. Oktober 2017")

    def test_rendering_unicode(self):
        context = {"foo": "äüßéô"}
        template_string = "{{ foo }}"
        output = self.render_template(template_string, context)
        self.assertEqual(output, "äüßéô")

    def test_escape(self):
        template_string = "{{ value | latex_escape }}"
        parameters = [
            ("&", "\\&"),
            ("%", "\\%"),
            ("$", "\\$"),
            ("#", "\\#"),
            ("_", "\\_"),
            ("{", "\\{"),
            ("}", "\\}"),
            ("~", "\\textasciitilde{}"),
            ("^", "\\textasciicircum{}"),
            ("\\", "\\textbackslash{}"),
            ("\\\\", "\\textbackslash{}\\textbackslash{}"),
            ("foo", "foo"),
        ]
        for value, expected in parameters:
            with self.subTest(value):
                output = self.render_template(template_string, {"value": value})
                self.assertEqual(output, expected)

    def test_linebreaks(self):
        context = {
            "brecht": "Ich sitze am Straßenhang." + "\nDer Fahrer wechselt das Rad."
        }
        template_string = "{{ brecht | linebreaks }}"
        output = self.render_template(template_string, context)
        self.assertEqual(
            output, r"Ich sitze am Straßenhang.\\" + "\nDer Fahrer wechselt das Rad."
        )
        # Render with default django renderer
        output = self.render_template(template_string, context, using="django")
        self.assertHTMLEqual(
            output,
            "<p>Ich sitze am Straßenhang.<br>" + "Der Fahrer wechselt das Rad.</p>",
        )

    @override_settings(
        TEMPLATES=[
            {
                "NAME": "tex",
                "BACKEND": "django_tex.engine.TeXEngine",
                "OPTIONS": {"environment": "tests.environment.test_environment"},
            }
        ]
    )
    def test_custom_filters(self):
        context = {
            "duration": datetime.timedelta(minutes=90),
        }
        template_string = "{{ duration | hhmm_format }}"
        output = self.render_template(template_string, context)
        self.assertEqual("1:30", output)

    @override_settings(LATEX_GRAPHICSPATH=["c:\\foo\\bar", "c:\\bar baz\\foo"])
    def test_graphicspath(self):
        template_string = "{% graphicspath %}"
        with override_settings(LATEX_INTERPRETER="pdflatex"):
            output = self.render_template(template_string)
            self.assertEqual(
                output, '\\graphicspath{ {c:/foo/bar/} {"c:/bar baz/foo/"} }'
            )
        with override_settings(LATEX_INTERPRETER="lualatex"):
            output = self.render_template(template_string)
            self.assertEqual(
                output, "\\graphicspath{ {c:/foo/bar/} {c:/bar baz/foo/} }"
            )


class Models(TestCase):
    """
    TeXTemplateFile contains the relative path to a tex template (e.g. django_tex/test.tex)
    and validates if this template can be loaded.abs

    Since TeXTemplateFile is an abstract base class, it is used here in a subclassed version 'TemplateFile'
    """

    def test_validation(self):
        TemplateFile(title="valid", name="tests/test.tex").full_clean()

        with self.assertRaises(ValidationError):
            TemplateFile(title="invalid", name="template/doesnt.exist").full_clean()


class Views(TestCase):
    def test_render_to_pdf(self):
        request = None  # request is only needed to make the signature of render_to_pdf similar to the signature of django's render function
        template_name = "tests/test.tex"
        context = {
            "test": "a simple test",
            "number": Decimal("1000.10"),
            "date": datetime.date(2017, 10, 25),
            "names": ["Arjen", "Robert", "Mats"],
        }
        response = render_to_pdf(request, template_name, context, filename="test.pdf")
        self.assertIsInstance(response, HttpResponse)
        self.assertEqual(response["Content-Type"], "application/pdf")
        self.assertEqual(response["Content-Disposition"], 'filename="test.pdf"')



import warnings


def render_to_pdf(*args, **kwargs):
    from django_tex.shortcuts import render_to_pdf as func

    warnings.warn(
        (
            '"django_tex.views.render_to_pdf" is deprecated '
            "and will be removed in future releases. Please use "
            '"django_tex.shortcuts.render_to_pdf" instead.'
        ),
        PendingDeprecationWarning,
    )
    return func(*args, **kwargs)



from django_tex.environment import environment


def hhmm_format(value):
    total_seconds = value.total_seconds()
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return "{:n}:{:02n}".format(hours, minutes)


def test_environment(**options):
    env = environment(**options)
    env.filters.update({"hhmm_format": hhmm_format})
    return env



from django.conf import settings
from jinja2.ext import Extension
from jinja2 import nodes


def format_path_for_latex(path):
    path = path.replace("\\", "/")
    if not path.endswith("/"):
        path += "/"
    if getattr(settings, "LATEX_INTERPRETER", None) != "lualatex" and " " in path:
        path = '"' + path + '"'
    path = "{" + path + "}"
    return path


class GraphicspathExtension(Extension):
    """Adds a `graphicspath` tag to Jinja2 that
    prints out a \\graphicspath{ {<path>} } command, where
    <path> is derived from the LATEX_GRAPHICSPATH setting or
    the BASE_DIR setting by default.
    """

    tags = set(["graphicspath"])

    def parse(self, parser):
        list_of_paths = getattr(settings, "LATEX_GRAPHICSPATH", [settings.BASE_DIR])
        value = (
            "\\graphicspath{ "
            + " ".join(map(format_path_for_latex, list_of_paths))
            + " }"
        )
        node = nodes.Output(lineno=next(parser.stream).lineno)
        node.nodes = [nodes.MarkSafe(nodes.Const(value))]
        return node



"""
Django settings for running tests
"""
import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

SECRET_KEY = "secret"

INSTALLED_APPS = [
    "django_tex",
    "tests",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
    {
        "NAME": "tex",
        "BACKEND": "django_tex.engine.TeXEngine",
        "APP_DIRS": True,
        "DIRS": [os.path.join(BASE_DIR, "custom_templates")],
    },
]

LANGUAGE_CODE = "de-de"

TIME_ZONE = "Europe/Berlin"

USE_I18N = True

USE_L10N = True

USE_TZ = True

DEBUG = True



from jinja2 import Environment

from django_tex.filters import FILTERS


def environment(**options):
    options.update(
        {
            "autoescape": None,
            "extensions": ["django_tex.extensions.GraphicspathExtension"],
        }
    )
    env = Environment(**options)
    env.filters = FILTERS
    return env



from django.template.backends.jinja2 import Jinja2

DEFAULT_ENVIRONMENT = "django_tex.environment.environment"


class TeXEngine(Jinja2):
    app_dirname = "templates"

    def __init__(self, params):
        if (
            "environment" not in params["OPTIONS"]
            or not params["OPTIONS"]["environment"]
        ):
            params["OPTIONS"]["environment"] = DEFAULT_ENVIRONMENT
        super().__init__(params)



from django.urls import path
import datetime
from django_tex.shortcuts import render_to_pdf


def index(request):
    context = {
        "title": "Die erste Mannschaft des VfL in der Saison 2020/2021",
        "number": 3.14,
        "date": datetime.date.today(),
        "names": [
            "Manuel Riemann",
            "Patrick Drewes",
            "Paul Grave",
            "Cristian Gamboa",
            "Danilo Soares",
            "Saulo Decarli",
            "Herbert Bockhorn",
            "Vasilios Lampropoulos",
            "Maxim Leitsch",
            "Armel Bella Kotchap",
            "Erhan Masovic",
            "Anthony Losilla",
            "Thomas Eisfeld",
            "Raman Chibsah",
            "Robert Tesche",
            "Robert Zulj",
            "Simon Zoller",
            "Tom Weilandt",
            "Soma Novothny",
            "Danny Blum",
            "Tarsis Bonga",
            "Gerrit Holtmann",
            "Milos Pantovic",
            "Baris Ekincier",
            "Silvère Ganvoula",
        ],
        "special_characters": "&$%#_{}",
    }
    return render_to_pdf(request, "tests/index.tex", context)


urlpatterns = [
    path("", index, name="index"),
]



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="django-tex",
    description="A simple Django app to render Latex templates and compile them into PDF files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/weinbusch/django-tex",
    author="Martin Bierbaum",
    license="MIT",
    keywords="django latex jinja2",
    packages=find_packages(exclude=["tests"]),
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=["django>=2.2", "jinja2>=2.9"],
    python_requires=">=3.6.2",
    package_data={"": ["*.tex"]},
)



import os
from subprocess import PIPE, run, CalledProcessError
import tempfile

from django.template.loader import get_template

from django_tex.exceptions import TexError
from django.conf import settings

DEFAULT_INTERPRETER = "lualatex"


def run_tex(source, template_name=None):
    with tempfile.TemporaryDirectory() as tempdir:
        return run_tex_in_directory(source, tempdir, template_name=template_name)


def run_tex_in_directory(source, directory, template_name=None):
    filename = "texput.tex"
    command = getattr(settings, "LATEX_INTERPRETER", DEFAULT_INTERPRETER)
    latex_interpreter_options = getattr(settings, "LATEX_INTERPRETER_OPTIONS", "")
    with open(os.path.join(directory, filename), "x", encoding="utf-8") as f:
        f.write(source)
    args = f"{command} -interaction=batchmode {latex_interpreter_options} {filename}"
    try:
        run(
            args,
            shell=True,
            stdout=PIPE,
            stderr=PIPE,
            check=True,
            cwd=directory,
        )
    except CalledProcessError as called_process_error:
        try:
            with open(
                os.path.join(directory, "texput.log"), "r", encoding="utf-8"
            ) as f:
                log = f.read()
        except FileNotFoundError:
            raise called_process_error
        else:
            raise TexError(log=log, source=source, template_name=template_name)
    with open(os.path.join(directory, "texput.pdf"), "rb") as f:
        pdf = f.read()
    return pdf


def compile_template_to_pdf(template_name, context):
    source = render_template_with_context(template_name, context)
    return run_tex(source, template_name=template_name)


def render_template_with_context(template_name, context):
    template = get_template(template_name, using="tex")
    return template.render(context)



import re


error_patterns = [
    r"^\!.*?l\.(?P<lineno>\d+).*?$",
    r"^\! Emergency stop.*?\*{3}.*?$",
    r"^\!.*?$",
]


ERROR = re.compile(r"|".join(error_patterns), re.DOTALL + re.MULTILINE)


class TexError(Exception):
    def __init__(self, log, source, template_name=None):
        self.log = log
        self.source = source.splitlines()

        mo = ERROR.search(self.log)

        self.message = mo.group() or "No error message found."

        if mo.group("lineno"):
            lineno = int(mo.group("lineno")) - 1
            total = len(self.source)
            top = max(0, lineno - 5)
            bottom = min(lineno + 5, total)
            source_lines = list(enumerate(self.source[top:bottom], top + 1))
            line, during = source_lines[lineno - top]

            self.template_debug = {
                "name": template_name,
                "message": self.message,
                "source_lines": source_lines,
                "line": line,
                "before": "",
                "during": during,
                "after": "",
                "total": total,
                "top": top,
                "bottom": bottom,
            }

            width = len(str(bottom + 1))

            template_context = "\n".join(
                "{lineno:>{width}} {line}".format(lineno=lineno, width=width, line=line)
                for lineno, line in source_lines
            )

            self.message += "\n\n" + template_context

    def __str__(self):
        return self.message
